import 'dart:mirrors';

String symbolToName(Symbol symbol) {
  final bs = symbol.toString();
  return bs.substring(8, bs.length - 2);
}

String nameUnderLine(String original) {
  return original
      .split('')
      .map((w) {
        if (w.toLowerCase() == w) {
          return w;
        } else {
          return '_${w.toLowerCase()}';
        }
      })
      .toList()
      .join('');
}

class Mapper<M> {
  Mapper();

  M? object(Map<String, dynamic> map) {
    return mapToObject(M, map);
  }

  dynamic mapToObject(Type cls, Map<String, dynamic> map) {
    final cm = reflectClass(cls);
    final params = <Symbol, dynamic>{};
    final listSeqs = <Symbol, MethodMirror>{};
    cm.instanceMembers.forEach((Symbol key, MethodMirror value) {
      if (value.isSynthetic && value.isGetter) {
        final mapKey = nameUnderLine(symbolToName(key));
        final fieldValue = map[mapKey];
        if ([int, double, bool, String]
            .contains(value.returnType.reflectedType)) {
          if (map.containsKey(mapKey)) {
            if (fieldValue != null) {
              params[key] = fieldValue;
            }
          }
        } else {
          if (value.returnType.reflectedType.toString().startsWith('List')) {
            listSeqs[key] = value;
          } else {
            if (map.containsKey(mapKey)) {
              if (fieldValue != null) {
                final objMap =
                    mapToObject(value.returnType.reflectedType, fieldValue);
                params[key] = objMap;
              }
            }
          }
        }
      }
    });
    final ins = cm.newInstance(const Symbol(''), [], params);
    final returnObject = ins.reflectee;
    listSeqs.forEach((Symbol key, MethodMirror value) {
      final mapKey = nameUnderLine(symbolToName(key));
      final fieldValue = map[mapKey];
      final om = reflect(returnObject);
      final listField = om.getField(key);
      final subcm = value.returnType.typeArguments.first;
      (fieldValue as List).forEach((ele) {
        if ([int, double, bool, String].contains(subcm.reflectedType)) {
          listField.delegate(Invocation.method(Symbol('add'), [ele]));
        } else {
          final subobj = mapToObject(subcm.reflectedType, ele);
          listField.delegate(Invocation.method(Symbol('add'), [subobj]));
        }
      });
    });
    return returnObject;
  }

  Map<String, dynamic> toMap(M object) {
    return objectToMap(M, object);
  }

  Map<String, dynamic> objectToMap(Type cls, dynamic object) {
    final om = reflect(object);
    final cm = reflectClass(cls);
    final map = <String, dynamic>{};
    cm.instanceMembers.forEach((Symbol key, MethodMirror value) {
      if (value.isSynthetic && value.isGetter) {
        final mapKey = nameUnderLine(symbolToName(key));
        if ([int, double, bool, String]
            .contains(value.returnType.reflectedType)) {
          map[mapKey] = om.getField(key).reflectee;
        } else {
          if (value.returnType.reflectedType.toString().startsWith('List')) {
            final list = om.getField(key).reflectee;
            final subcm = value.returnType.typeArguments.first;
            final listIns = [];
            (list as List).forEach((subObj) {
              listIns.add(objectToMap(subcm.reflectedType, subObj));
            });
            map[mapKey] = listIns;
          } else {
            print('cm:$cm key: $key');
            final subObj = om.getField(key).reflectee;
            final subMap = objectToMap(value.returnType.reflectedType, subObj);
            map[mapKey] = subMap;
          }
        }
      }
    });
    return map;
  }
}
